unit uYouTube;
{$DEFINE OUTPUTSTREAM}

interface

uses
  SysUtils, Classes, HttpSend, blcksock, RegExpr;

type
  TStringEvent = procedure(Sender: TObject; var Value: string) of object;
  TProgressEvent = procedure(Sender: TObject; ID: integer; const Data: string; Progress, Max: int64; var DoAbort: boolean) of object;

  TYouTubeProgress = (ytpStarting,
                      ytpDownloadingVideoPage,
                      ytpGettingTitle, ytpFoundTitle,
                      ytpGettingFileName, ytpFoundFileName,
                      ytpGettingParams, ytpFoundTimestamp, ytpFoundVideoID,
                      ytpDownloadingVideo,
                      ytpFinished,
                      ytpDownloadError, ytpParamsError);

  TYouTube = class
    private
      fOnProgress: TProgressEvent;
      fOnTitle: TStringEvent;
      fOnFileName: TStringEvent;
      fHttp: THttpSend;
      fTitleRE: TRegExpr;
      fParamsRE: TRegExpr;
    protected
      fLastGoodStatus: THookSocketReason;
      BytesTransferred: int64;
      Video: THttpSend;
      VideoUrl: string;
      function DoTitle(const Title: string): string; virtual;
      function DoFileName(const FileName: string): string; virtual;
      function DownloadPage(Http: THttpSend; Url: string): boolean; overload; virtual;
      function DownloadPage(Http: THttpSend; Url: string; out Page: string): boolean; overload; virtual;
      procedure DoProgress(Step: TYouTubeProgress; const Data: string; Progress, Max: int64); virtual;
      procedure VideoSockStatus(Sender: TObject; Reason: THookSocketReason; const Value: string); virtual;
      function CreateHttp: THttpSend; virtual;
      property TitleRE: TRegExpr read fTitleRE;
      property ParamsRE: TRegExpr read fParamsRE;
    public
      constructor Create; virtual;
      destructor Destroy; override;
      function Download(const Url: string): boolean; virtual;
      property OnProgress: TProgressEvent read fOnProgress write fOnProgress;
      property OnTitle: TStringEvent read fOnTitle write fOnTitle;
      property OnFileName: TStringEvent read fOnFileName write fOnFileName;
      property DefaultHttp: THttpSend read fHttp;
      property LastGoodStatus: THookSocketReason read fLastGoodStatus;
    end;

const LastGoodStatusStr: array[THookSocketReason] of string
                       = ('Resolving URL.', 'URL resolved.', 'Creating socket.', 'Socket created.', 'Binding socket.',
                          'Connecting to site.', 'Testing incoming data.', 'Testing outgoing data.',
                          'Listening socket.', 'Accepting connection.', 'Reading data.', 'Writing data.',
                          'Waiting for data.', 'Unknown error.');

implementation

uses
  ulkJSON;

function StrTr(const Kde, Co, Cim: string): string;
var i, j: integer;
begin
  Result := Kde;
  if Kde <> '' then
    for i := 1 to Length(Co) do
      repeat
        j := Pos(Co[i], Result);
        if j > 0 then
          Result[j] := Cim[i];
      until j <= 0;
end;

{ TYouTube }

const TITLERE_TITLEINDEX = 2;
      PARAMSRE_JSONINDEX = 1;
      
constructor TYouTube.Create;
begin
  inherited Create;
  fHttp := THttpSend.Create;
  fTitleRE := TRegExpr.Create;
  fTitleRE.ModifierI := True;
  fTitleRE.ModifierM := True;
  fTitleRE.Expression := '<title>(YouTube - )?(.+?)</title>';
  fParamsRE := TRegExpr.Create;
  fParamsRE.ModifierI := True;
  fParamsRE.ModifierM := True;
  fParamsRE.Expression := '\sswfArgs\s*=\s*(\{.*?\})';
end;

destructor TYouTube.Destroy;
begin
  FreeAndNil(fHttp);
  FreeAndNil(fTitleRE);
  FreeAndNil(fParamsRE);
  inherited;
end;

function TYouTube.CreateHttp: THttpSend;
begin
  Result := THttpSend.Create;
  Result.UserAgent := DefaultHttp.UserAgent;
  Result.ProxyHost := DefaultHttp.ProxyHost;
  Result.ProxyPort := DefaultHttp.ProxyPort;
  Result.ProxyUser := DefaultHttp.ProxyUser;
  Result.ProxyPass := DefaultHttp.ProxyPass;
end;

function TYouTube.DoTitle(const Title: string): string;
begin
  Result := Title;
  if Assigned(OnTitle) then
    OnTitle(Self, Result);
end;

function TYouTube.DoFileName(const FileName: string): string;
begin
  Result := FileName;
  if Assigned(OnFileName) then
    OnFileName(Self, Result);
end;

procedure TYouTube.DoProgress(Step: TYouTubeProgress; const Data: string; Progress, Max: int64);
var DoAbort: boolean;
begin
  DoAbort := False;
  if Assigned(OnProgress) then
    OnProgress(Self, Integer(Step), Data, Progress, Max, DoAbort);
  if DoAbort then
    Abort;
end;

function TYouTube.DownloadPage(Http: THttpSend; Url: string): boolean;
const Location = 'Location:';
var Cont: boolean;
    i: integer;
begin
  repeat
    Http.Clear;
    Result := Http.HttpMethod('GET', Url);
    Cont := False;
    if Result and (Http.ResultCode >= 300) and (Http.ResultCode < 400) then
      for i := 0 to Pred(Http.Headers.Count) do
        if AnsiCompareText(Location, Copy(Http.Headers[i], 1, Length(Location))) = 0 then
          begin
          Url := Trim(Copy(Http.Headers[i], Length(Location)+1, MaxInt));
          Cont := True;
          Break;
          end;
  until not Cont;
end;

function TYouTube.DownloadPage(Http: THttpSend; Url: string; out Page: string): boolean;
begin
  Page := '';
  Result := DownloadPage(Http, Url);
  if Result then
    begin
    SetLength(Page, Http.Document.Size);
    Http.Document.Seek(0, 0);
    Http.Document.ReadBuffer(Page[1], Http.Document.Size);
    end;
end;

function TYouTube.Download(const Url: string): boolean;
var Info: THttpSend;
    Page, Title, FileName, Timestamp, VideoID: string;
    JSON: TlkJSONObject;
begin
  Result := True;
  try
    DoProgress(ytpStarting, Url, 0, 1);
    Video := nil;
    VideoUrl := '';
    DoProgress(ytpDownloadingVideoPage, Url, 0, 1);
    Info := CreateHttp;
    try
      if DownloadPage(Info, Url, Page) then
        begin
        DoProgress(ytpGettingTitle, Page, 0, Length(Page));
        if TitleRE.Exec(Page) then
          Title := DoTitle(TitleRE.Match[TITLERE_TITLEINDEX])
        else
          Title := DoTitle('');
        DoProgress(ytpFoundTitle, Title, 0, 1);
        DoProgress(ytpGettingFileName, Page, 0, Length(Page));
        FileName := DoFileName(StrTr(Title, '\/:*?"<>|', '--;..''--!') + '.flv');
        DoProgress(ytpFoundFileName, FileName, 0, 1);
        DoProgress(ytpGettingParams, Page, 0, Length(Page));
        Timestamp := '';
        VideoID := '';
        if ParamsRE.Exec(Page) then
          begin
          JSON := TlkJSON.ParseText(ParamsRE.Match[PARAMSRE_JSONINDEX]) as TlkJSONobject;
          try
            Timestamp := Trim(JSON.getString('t'));
            VideoID := Trim(JSON.getString('video_id'));
          finally
            JSON.Free;
          end;
        if Timestamp <> '' then
          DoProgress(ytpFoundTimestamp, Timestamp, 0, 1);
        if VideoID <> '' then
          DoProgress(ytpFoundVideoID, VideoID, 0, 1);
        if (TimeStamp <> '') and (VideoID <> '') then
          begin
          VideoURL := 'http://www.youtube.com/get_video.php?video_id=' + VideoID + '&t=' + TimeStamp;
          DoProgress(ytpDownloadingVideo, VideoURL, 0, 1);
          Video := CreateHttp;
          try
            Video.Cookies.Assign(Info.Cookies);
            {$IFDEF OUTPUTSTREAM}
            Video.OutputStream := TFileStream.Create(FileName, fmCreate);
            try
              try
            {$ENDIF}
                Video.Sock.OnStatus := VideoSockStatus;
                BytesTransferred := 0;
                fLastGoodStatus := HR_Error;
                if DownloadPage(Video, VideoURL) then
                  begin
                  {$IFNDEF OUTPUTSTREAM}
                  with TFileStream.Create(FileName, fmCreate) do
                    try
                      Video.Document.Seek(0, 0);
                      CopyFrom(Video.Document, 0);
                    finally
                      Free;
                      end;
                  {$ENDIF}
                  DoProgress(ytpFinished, FileName, 1, 1);
                  end
                else
                  DoProgress(ytpDownloadError, Video.ResultString, 1, 1);
            {$IFDEF OUTPUTSTREAM}
              finally
                Video.OutputStream.Free;
                Video.OutputStream := nil;
                end;
            except
              on Exception do
                begin
                if FileExists(FileName) then
                  DeleteFile(FileName);
                Raise;
                end;
              end;
            {$ENDIF}
          finally
            Video.Free;
            Video := nil;
            VideoUrl := '';
            end;
          end
        else
          DoProgress(ytpParamsError, Page, 1, 1);
        end;
      end
    else
      DoProgress(ytpDownloadError, Info.ResultString, 1, 1);
    finally
      Info.Free;
      end;
  except
    on EAbort do
      begin
      Result := False;
      end;
    end;
end;

procedure TYouTube.VideoSockStatus(Sender: TObject; Reason: THookSocketReason; const Value: string);
begin
  if Reason <> HR_Error then
    fLastGoodStatus := Reason;
  if (Reason = HR_ReadCount) then
    begin
    BytesTransferred := BytesTransferred + StrToInt64(Value);
    DoProgress(ytpDownloadingVideo, VideoUrl, BytesTransferred, Video.DownloadSize);
    end;
end;

end.
